#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "lion.h"
#include "misc.h"
#include "lfnmatch.h"
#include "dirlist.h"

#define DEBUG_FLAG 8
#include "debug.h"
#include "llrar.h"
#include "conf.h"
#include "parser.h"
#include "request.h"
#include "skin.h"


enum llrar_unrar_enum {
	LLRAR_UNRAR_NONE = 0,
	LLRAR_UNRAR_NORMAL,
	LLRAR_UNRAR_SEEK,
};
typedef enum llrar_unrar_enum llrar_unrar_t;


static int llrar_handler(lion_t *handle, void *user_data, 
						 int status, int size, char *line);
static int llrar_extract_handler(lion_t *handle, void *user_data, 
								 int status, int size, char *line);

static void llrar_test_unrar();


static int llrar_spawned_list = 0;
static int llrar_spawned_get  = 0;
static int llrar_resumed_get  = 0;

static llrar_unrar_t llrar_unrar = LLRAR_UNRAR_NONE;

// How many bytes should the seek be before we switch to re-spawning
// using the seek option.
static lion64u_t LLRAR_SEEK_TRIGGER = 100 * 1024 * 1024;


int llrar_init(void)
{
	// We spawn a unrar process here, and check if it has the special 
	// option to seek on stored rars. If so, we flag this support.

	llrar_test_unrar();

	return 0;
}



void llrar_free(void)
{

	debugf("[llrar] spawned %3d unrar list processes.\n"
		   "[llrar] spawned %3d unrar  get processes.\n"
		   "[llrar] resumed %3d unrar  get processes.\n",
		   llrar_spawned_list,
		   llrar_spawned_get,
		   llrar_resumed_get);
}


//
// This function is called a lot.
//
void llrar_compute_blocksize(request_t *node)
{
	lion64_t remaining, buffer, before;

	// If we are asking for just a part of the file, we need to work out
	// how much more they want, and if that is smaller than the next 
	// buffer read, we set the buffer size smaller.
	if (!node->bytes_to && !node->bytes_from)
		return;

	buffer = (lion64_t) lion_get_buffersize(node->althandle);
	before = buffer;

	// lseek, are we before from, and would move past it?
	if (node->bytes_from) {

		if ((node->bytes_sent < node->bytes_from) &&
			((node->bytes_sent + buffer) > node->bytes_from)) {
			buffer = node->bytes_from - node->bytes_sent;

			debugf("[llrar] adjusting buffersize from %"PRIu64".\n",
				   buffer);
		}

	}


	if (node->bytes_to) {

		if (node->bytes_sent > node->bytes_to) {
			debugf("[llrar] bug, already sent > to\n");
			return;
		}

		remaining = node->bytes_to - node->bytes_sent + 1;
		
		if (buffer > remaining) {
			buffer = remaining;

			debugf("[llrar] adjusting buffersize to %"PRIu64".\n",
				   buffer);
		}
	}


	
	if (before != buffer)
		lion_set_buffersize(node->althandle, (int)buffer);

}







void llrar_stop_process(request_t *node)
{

	// No process? Just return
	if (!node->althandle)
		return;

	debugf("[llrar] stopping old process at %"PRIu64"\n", node->bytes_sent);


	// We need to kill the process without fiddling with anything else really.
	// userdata to NULL, then all handlers do nothing.
	lion_set_userdata(node->althandle, NULL);

	// Allow it to drain
	lion_enable_read(node->althandle);

	// Disconnect it
	lion_disconnect(node->althandle);

	// Set it to NULL
	node->althandle = NULL;

	SAFE_FREE(node->rar_name);
	node->bytes_sent = 0;


}



//
// Take a http request node, which should have "cwd" filled in appropriately.
// We then call dirlist to issue listing(s) for this path.
//
void llrar_list(request_t *node, char *executable)
{
	char buffer[1024];

	debugf("[llrar] file listing requested: %s - %s\n",
		   node->disk_path, node->path);

	if (llrar_unrar == LLRAR_UNRAR_NONE) {
		request_reply(node, 500, "No unrar program found");
		return;
	}



	// If we already have a process running, kill it
	llrar_stop_process(node);


	SAFE_FREE(node->rar_name);

	// Spawn unrar
	//-ap$DIR -av- -c- -p- -y
	// ok, -ap$DIR does not work in listings.
	// but -ndirectory1/ -ndirectory1/directory2/ does appear to work. Needs
	// trailing slash, or it shows itself.
#if 1
	// Old style was to use "-n/path" to attempt to list only one directory 
	// within a .rar, but unrar has a bug where it will just not show the
	// directory names, etc. So, we change to using "v" to list. 
	// Now we get one line with full path, followed by a line of details.
	// That way, we can skip the first space, then look for last "/", this is
	// the full path, and filenames. Spaces included.
	// 
    //  directory1/directory2/file2.avi
    //          7       17 242% 04-12-07 14:11 -rw-r--r-- 16B28489 m3b 2.9
	//
	// We then filter out any entries for the path we care about, drop the
	// rest. No slash in filename means it is in root.

	snprintf(buffer, sizeof(buffer), "%s v -v -c- -p- -y -cfg- -- \"%s\"",
			 executable,
			 node->disk_path);


#endif


	node->rar_name = strdup("rar listing");

	
	
	debugf("[llrar] spawning new '%s'\n", buffer);
			
	node->unrar_parse = 0;
	
	llrar_spawned_list++;

	node->althandle = lion_system(buffer, 1, LION_FLAG_NONE, node);

	if (!node->althandle) {
		debugf("[llrar] failed to launch unrar\n");
		request_reply(node, 513, "failed to launch unrar");
		return;
	}
	
	// Set list handler
	lion_set_handler(node->althandle, llrar_handler);
	
}



void llrar_get(request_t *node, char *executable)
{
	char buffer[1024];
	int keep_process;

	if (llrar_unrar == LLRAR_UNRAR_NONE) {
		request_reply(node, 500, "No unrar program found");
		return;
	}

	debugf("[llrar] get file requested\n");

	// Alas, subtitle lookup will have wrong size:(
    if (node->rar_file && !lfnmatch("*.srt",
									node->rar_file,
									LFNM_CASEFOLD)) {
		node->bytes_size = 500 * 1024;
    }


	// HEAD MODE, DO NOTHING
	if (node->head_method)
		return;

	// Safety, we aren't called if this is not set in request.c
	if (!node->rar_file)
		return;
	
	// Now, we both need to url_decode this part, and re-assign.
	//	name = request_url_decode(ar);

	// Keep process?
	keep_process = 0;


	debugf("[llrar] checking for resume: rar_name (%s), althandle (%p), name (%s), from (%"PRIu64") >= sent (%"PRIu64")\n",
		   node->rar_name ? node->rar_name : "",
		   node->althandle,
		   node->rar_file ? node->rar_file : "",
		   node->bytes_from,
		   node->bytes_sent);


	if (node->rar_name &&
		node->althandle &&
		!mystrccmp(node->rar_file, node->rar_name) &&
		node->bytes_from >= node->bytes_sent) {
		keep_process = 1;
		debugf("[llrar] resuming old unrar\n");
		llrar_resumed_get++;
	}


	// Now, if we have special unrar with seek support, we should use it if
	// we are seeking far.
	if ((llrar_unrar == LLRAR_UNRAR_SEEK) &&
		keep_process) {
		// We thought we were going to keep the process, but lets see if the
		// seek is too big, we might as well abort, and use the seek feature.

		if ((node->bytes_from - node->bytes_sent) >=
			LLRAR_SEEK_TRIGGER) {
			keep_process = 0;
			
			debugf("[llrar] seek too large, respawning with seek: %"PRIu64" <= %"PRIu64"\n",
				   node->bytes_from,
				   node->bytes_sent);
		}
	}

	if (!keep_process) {


		llrar_stop_process(node);

		// Copy the rar_file name to rar_name, so we can compare for resumes
		node->rar_name = strdup(node->rar_file);


		if ((llrar_unrar == LLRAR_UNRAR_NORMAL) || !node->bytes_from) {
			
			snprintf(buffer, sizeof(buffer), 
					 "%s p -inul -c- -p- -y -cfg- -- \"%s\" \"%s\"",
					 executable,
					 node->disk_path,
					 (node->rar_name[0] == '/') ? &node->rar_name[1] : 
					 node->rar_name);

		} else {

			debugf("[[llrar] using unrar with seek %"PRIu64".\n",
				   node->bytes_from);

			snprintf(buffer, sizeof(buffer), 
					 "%s p -inul -c- -p- -y -cfg- -sk%"PRIu64" -- \"%s\" \"%s\"",
					 executable,
					 node->bytes_from,
					 node->disk_path,
					 (node->rar_name[0] == '/') ? &node->rar_name[1] : 
					 node->rar_name);

			// Make it think we've already seen "seek" bytes.
			node->bytes_sent = node->bytes_from;

		}
	
		debugf("[llrar] spawning new '%s'\n", buffer);
			
		node->unrar_parse = 0;
	
		llrar_spawned_get++;
	
		node->althandle = lion_system(buffer, 1, LION_FLAG_NONE, node);

		if (!node->althandle) {
			debugf("[llrar] failed to launch unrar\n");
			request_reply(node, 513, "failed to launch unrar");
			return;
		}

		// Set extract handler
		lion_set_handler(node->althandle, llrar_extract_handler);
		// Set binary mode
		lion_enable_binary(node->althandle);
		// Disable read until we are ready
		lion_disable_read(node->althandle);


	}  // !keep_process
	
	
	// Set right chunk mode.
	llrar_compute_blocksize(node);

	debugf("[llrar] processing: sent (%"PRIu64") bytes_from (%"PRIu64") bytes_to (%"PRIu64")\n",
		   node->bytes_sent, node->bytes_from, node->bytes_to);

}










static int llrar_handler(lion_t *handle, void *user_data, 
						 int status, int size, char *line)
{
	request_t *node = (request_t *) user_data;
	int directory;
	char list[1024];
	char *name, *fsize, *packed, *ratio, *date, *thyme, *attr;
	char *ar, *path, *slash, *root;

	// If node isn't set, it's b0rken.
	if (!node) return 0;

	switch(status) {

	case LION_CONNECTION_CONNECTED:
		debugf("[llrar] connected to external process, issuing root for /%s\n",
			   node->path);
		break;


	case LION_PIPE_RUNNING:
		debugf("[llrar] unrar is running\n");
		//lion_printf(handle, "HELLO THERE\r\n");
		break;
	case LION_PIPE_FAILED:
		debugf("[llrar] unrar failed to launch: %d:%s\n", size, line);
		node->althandle = NULL;
		SAFE_FREE(node->rar_name);
		request_finishdir(node);
		break;
	case LION_PIPE_EXIT:
		debugf("[llrar] unrar finished: %d\n", size);
		if (size == -1) {
			lion_want_returncode(handle);
			break;
		}
		node->althandle = NULL;
		SAFE_FREE(node->rar_name);
		request_finishdir(node);
		break;

	case LION_INPUT:
// Name             Size   Packed Ratio  Date   Time     Attr      CRC   Meth Ver
//-------------------------------------------------------------------------------
// aaf-rc.s03e13.avi 105119744 14999906  --> 11-11-07 23:52  .....A.   78AF2535 m0g 2.0
//-------------------------------------------------------------------------------
		debugf("[llrar] >> '%s'\n", line);

		switch(node->unrar_parse) {
		case 0:  // start of unrar list, waiting for line with -----
			if (!strncmp("--------------------", line, 20))
				node->unrar_parse++;
			break;

		case 1:  // woohooo, actually parsing entries
			// This is the filename part
			// " directory1/directory2/file2.avi"

			if (!strncmp("--------------------", line, 20)) {
				node->unrar_parse = 0;
				break;
			}

			if (*line != ' ') {
				debugf("[llrar] unable to parse: '%s'\n", line);
				break;
			}

			// Skip that leading space.
			if (*line == ' ')
				line++;

			// Remember this line until next parsing line.
			node->rar_file = strdup(line);

			node->unrar_parse++;
			break;

		case 2: // parse out filesize, type etc.
			// "   7       17 242% 04-12-07 14:11 -rw-r--r-- 16B28489 m3b 2.9"

			// Alternate filename/data lines, change back to filename
			node->unrar_parse = 1;

			ar = line;
			
			// node->rar_name hold the FULL path name.
			fsize  = misc_digtoken(&ar, " \r\n");
			packed = misc_digtoken(&ar, " \r\n");
			ratio  = misc_digtoken(&ar, " \r\n");
			date   = misc_digtoken(&ar, " \r\n");
			thyme  = misc_digtoken(&ar, " \r\n");
			attr   = misc_digtoken(&ar, " \r\n");
			
			if (!node->rar_file ||
				!fsize|| !*fsize||
				!attr || !*attr) {
				debugf("[llrar] unable to parse -- skipping\n");
				break;
			}

			// Files spanning Volumes will be repeated, but we can tell by
			// lookip at "ratio" column.
			if (!mystrccmp("<->", ratio) ||
				!mystrccmp("<--", ratio)) {

				SAFE_FREE(node->rar_file);
				break;
			}


			debugf("[llrar] parsed name '%s'\n", node->rar_file);
			
			// Now we need to look at the rar_file fullname, and 
			// split it into directory/filename 
			// Find last slash. Will windows use '\' ?
#ifdef WIN32
			for (name = node->rar_file; *name; name++)
				if (*name == '\\') *name = '/';
#endif
			slash = strrchr(node->rar_file, '/');
			if (!slash) { // We are in root.
				path = "/";
				name = node->rar_file;
			} else { // We are in a subdir
				*slash = 0;
				name = &slash[1];
				path = node->rar_file;
			}
			// Now "path" should hold full path, and "name" just the
			// entry name.
			if (node->rar_directory) {
				root = node->rar_directory;
				// Skip leading "/" as unrar wont start with /
				while (*root == '/') root++;
			} else
				root = "/";

			debugf("[llrar] Checking '%s' == '%s'\n",
				   path, root);


			//  Now check if it is for a path we care about
			if (!mystrccmp(path, root)) {
				// It is
				
				//rar attributes. If DOS format, it will be
				// .D.....    : 7 chars.
				// Unix format:
				// d--------- :  10 chars
				directory = 0;
				if ((tolower(*attr) == 'd') ||
					(tolower(attr[1]) == 'd')) 
					directory = 1;
				
				if (directory)
					snprintf(list, sizeof(list),
							 "drwxr-xr-x  1 unrar unrar %s Jun 7 %s %s&d=%s/%s",
							 fsize,
							 thyme,
							 node->path,
							 node->rar_directory ? node->rar_directory : "",
							 name);
				else
					snprintf(list, sizeof(list),
							 "-rwxr-xr-x  1 unrar unrar %s Jun 7 %s %s&f=%s,%s/%s",
							 fsize,
							 thyme,
							 node->path,
							 fsize,
							 node->rar_directory ? node->rar_directory : "",
							 name);

				skin_write_type(node, list);

			} // directory we care about (in our path)

			// Release the filename from previous line.
			SAFE_FREE(node->rar_file);

			break;

		case 3:  // seen ----- the rest is fluff
			break;
		}



				
		break;

	}

	return 0;
}








static int llrar_extract_handler(lion_t *handle, void *user_data, 
								 int status, int size, char *line)
{
	request_t *node = (request_t *) user_data;
	int toomuch;

	// If node isn't set, it's b0rken.
	if (!node) return 0;

	switch(status) {

	case LION_PIPE_RUNNING:
		debugf("[llrar] connected to unrar process\n");
		break;

	case LION_PIPE_FAILED:
		debugf("[llrar] unrar failed to launch: %d:%s\n", 
			   size, line ? line : "");

		node->althandle = NULL;
		SAFE_FREE(node->rar_name);
		request_reply(node, 500, line ? line : "Error with unrar");
		break;

	case LION_PIPE_EXIT:
		debugf("[llrar] unrar finished: %d\n", size);
		if (size == -1) {
			lion_want_returncode(handle);
			break;
		}
		node->althandle = NULL;
		SAFE_FREE(node->rar_name);

		debugf("[request] socket read restored\n");
		lion_enable_read(node->handle);

		if ((node->type & REQUEST_TYPE_CLOSED)) {
			debugf("[request] closing socket due to TYPE_CLOSED\n");
			lion_close(node->handle);
		}
		
		break;

	case LION_BINARY:

		// Check if we are lseeking
		if (node->bytes_from) {

			if ((node->bytes_sent + (lion64u_t)size) <= node->bytes_from) {
				if (!node->bytes_sent) {
					debugf("[llrar] seeking to start position\n");
				}
				// Just discard, do nothing.
				node->bytes_sent += (lion64u_t)size;
				// Compute new blocksize
				llrar_compute_blocksize(node);
				break;
			}

			// Complicated case, sent < from but sent + size > from
			if ((node->bytes_sent < node->bytes_from) &&
				(node->bytes_sent + (lion64u_t)size) > node->bytes_from) {
				debugf("[llrar] special case sent < from, but sent+size > from!\n");
				// decrease size, increase line
			}
		}


		// Check if we would send too much
		// Check if we have sent exactly enough
		toomuch = 0;

		
		if (node->bytes_to) {

			if ((node->bytes_sent + (lion64u_t) size) >
				node->bytes_to) 
				toomuch = 1;
		}

		node->bytes_sent += size;
		
		if (toomuch) {
			// bytes_sent is our bytes index into the unrar read file
			// but we need to make size smaller so that we don't send more
			// than promised.
			int before;

			before = size;
			size -= (int) node->bytes_sent - node->bytes_to - 1 ;
			
			debugf("[llrar] final block %d\n", size);
			if (before > size)
				debugf("[llrar] Warning, overshot our place, can't resume. read %d wanted %d\n",
					   before, size);

			// Sleep unrar process incase we want to read more soon
			// Warning, because we enable read of handle below
			// its handler enables althandle, so we must set it there
			lion_disable_read(node->althandle);
			
			// Set default buffersize again
			lion_set_buffersize(node->althandle, 0);

		}

		if (lion_output(node->handle, line, size) == -1) {
			debugf("[request] socket write failed.\n");
			break;
		}

		// Check if we should limit the next read
		if (!toomuch) {
			llrar_compute_blocksize(node);
		} else {

			if (node->handle) {
		  
				// Also close socket?
				// Warning, if this DOES close the socket, "node"
				// is entirely clear, and any references after is INVALID.
				if ((node->type & REQUEST_TYPE_CLOSED)) {
					debugf("[llrar] closing socket due to TYPE_CLOSED\n");
					// Close rar too
					lion_disconnect(node->althandle);
					// Close http socket
					lion_close(node->handle);
					break;
				}

				debugf("[llrar] socket read restored %p\n",
					   node->handle);
				lion_enable_read(node->handle);

			}
		}

		break;

	} // switch status

	return 0;
}






static int llrar_test_handler(lion_t *handle, void *user_data, 
							  int status, int size, char *line)
{
	char *ar, *opt;
	int *running = (int *) user_data;

	switch(status) {

	case LION_PIPE_RUNNING:
		debugf("[llrar] unrar is running\n");
		llrar_unrar = LLRAR_UNRAR_NORMAL; // We have unrar support
		break;
	case LION_PIPE_FAILED:
		debugf("[llrar] unrar failed to launch: %d:%s\n", size, line);
		*running = 0;
		break;
	case LION_PIPE_EXIT:
		debugf("[llrar] unrar finished: %d\n", size);
		if (size == -1) {
			lion_want_returncode(handle);
			break;
		}
		*running = 0;
		break;

	case LION_INPUT:
		//  sk<size>      Seek to byte offset size (-m0 archives only)
		if (line && *line) {
			//debugf("[llrar] >> '%s'\n", line);
			
			ar = line;
			opt = misc_digtoken(&ar, " \r\n");

			if (opt && *opt && !mystrccmp("sk<size>", opt)) {
				debugf("[llrar] upgrading to unrar with SEEK\n");
				llrar_unrar = LLRAR_UNRAR_SEEK; // We have unrar with seek support!
			}

		} // line
		break;

	} // switch

	return 0;
}



//
// There is a problem here, you can define a different unrar program
// for each TYPE line, and we should really test each one, and not assume
// globally.
//
void llrar_test_unrar(void)
{
	char *unrar;
	lion_t *tester;
	char buffer[1024];
	int running;

	llrar_unrar = LLRAR_UNRAR_NONE; // We have no unrar

	unrar = skin_get_rar();
	if (!unrar) return;

	
	snprintf(buffer, sizeof(buffer), "%s -c- -p- -y -cfg- -- ",
			 unrar);
	
	debugf("[llrar] spawning new '%s'\n", buffer);
			
	running = 1;
	
	tester = lion_system(buffer, 1, LION_FLAG_NONE, (void *)&running);

	if (!tester) {
		debugf("[llrar] failed to launch unrar\n");
		return;
	}
	
	// Set list handler
	lion_set_handler(tester, llrar_test_handler);

	//	while(running) {
	//	lion_poll(0,1);
	//};

}

