#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "lion.h"
#include "misc.h"
#include "lfnmatch.h"

#define DEBUG_FLAG 1024
#include "debug.h"
#include "conf.h"
#include "parser.h"
#include "request.h"
#include "visited.h"



static char *visited_dbfile = NULL;
static int   visited_mintime = 30;
static float visited_percent = 75.0;



// Because the way that Syabas gets videos, (open, lseek, read 16k, close)
// we never see "one long transfer" of media. This means we need to remember
// the last few transfers, and slowly tally up the byte count.

struct visited_struct {
	char *fname;
	bytes_t bytes;
	time_t ctime;
	time_t mtime;
	int seen;
};

typedef struct visited_struct visited_t;

#define VISITED_MAX 1000
THREAD_SAFE static visited_t visited_list[VISITED_MAX];

// Circular position index
THREAD_SAFE static int visited_insert_pos = 0; 
THREAD_SAFE static int visited_search_pos = 0; 
// Strictly not required, but speeds-up searches when less than MAX
THREAD_SAFE static int visited_num = 0;





int visited_init(void)
{
	int i;
	
	for (i = 0; i < VISITED_MAX; i++) {
		memset(&visited_list[i], 0, sizeof(visited_t));
	}

	visited_insert_pos = 0;
	visited_search_pos = 0;
	visited_num = 0;

	return 0;
}



void visited_free(void)
{

	int i;
	
	for (i = 0; i < VISITED_MAX; i++) {
		SAFE_FREE(visited_list[ i ].fname);
	}

	visited_insert_pos = 0;
	visited_search_pos = 0;
	visited_num = 0;

}





// 
//
// CMD      : VISITED
// Required : FILE, MINTIME, PERCENT
// Optional : 
//
void visited_cmd(char **keys, char **values,
			  int items,void *optarg)
{
	char *file, *mintime, *percent;

	// Required
	file    = parser_findkey(keys, values, items, "FILE");
	mintime = parser_findkey(keys, values, items, "MINTIME");
	percent = parser_findkey(keys, values, items, "PERCENT");

	if (!file || !*file || !mintime || !*mintime || !percent || !*percent) {
		printf("      : [visited] missing required field: FILE, MINTIME, PERCENT\n");
		//do_exit = 1;
		return;
	}

	SAFE_COPY(visited_dbfile, file);
	visited_mintime = atoi(mintime);
	visited_percent = atof(percent);

}



void visited_freenode(visited_t *vnode)
{
	SAFE_FREE(vnode->fname);
	memset(vnode, 0, sizeof(*vnode));
}




//
// Search for existing node matching this name, starting from 
// "search_pos". (Looping at the end). Update search_pos to found location.
//
visited_t *visited_find(char *fname) 
{
	int i, place;

	for (i = 0, place = visited_search_pos; 
		 i < visited_num; 
		 i++, place++) {

		//if (place >= VISITED_MAX) place = 0;
		if (place >= visited_num) place = 0;

		if (!mystrccmp(fname, visited_list[ place ].fname)) {
			visited_search_pos = place;
			return &visited_list[ place ];
		}
	}

	return NULL;
}



// Use the next "insert_pos" node, clear it, add in new name, and ctime.
// return it.
visited_t *visited_add(char *fname)
{
	visited_t *vnode;

	// Already added?
	if ((vnode = visited_find(fname))) return vnode;
		
	// insert_pos points to where we insert, free it first.
	vnode = &visited_list[ visited_insert_pos ];
	visited_freenode(vnode);

	// loop the list if we stepped over.
	if (visited_insert_pos++ >= VISITED_MAX) visited_insert_pos = 0;
	
	vnode->fname = strdup(fname);
	vnode->ctime = lion_global_time;
	
	if (visited_num < VISITED_MAX)
		visited_num++;

	debugf("[visited] added '%s'\n", fname);

	return vnode;
}



void visited_fileclosed(request_t *node)
{
	time_t duration;
	bytes_t in, out;
	float percent = 0.0;
	visited_t *vnode;

	// Quick reasons as to why we should ignore this file
	if (!node->althandle) return;
	if (!node->bytes_size) return;

	debugf("[visited] closed file '%s'\n", node->path);

	lion_get_bytes(node->althandle, &in, &out);

	// No bytes sent
	if (!in) return;

	duration = lion_get_duration(node->althandle);

	// Watched it for minimum time?
   	if (duration <= visited_mintime) return;

	// Calculate percentage bytes sent
	// look up the node, or create, so we can keep stats.
	vnode = visited_find(node->path);

	if (!vnode)
		vnode = visited_add(node->path);

	if (!node) return;


	// Update node.
	vnode->bytes += in;

	// Let's only update it once a second.
	if (vnode->mtime == lion_global_time) return;

	vnode->mtime = lion_global_time;


	//percent = node->bytes_size * 100 / vnode->bytes;
	percent = vnode->bytes * 100 / node->bytes_size;
	duration = vnode->mtime - vnode->ctime;

	debugf("[visited] file played for %ld seconds\n", duration);
	debugf("[visited] and we have sent %"PRIu64" bytes out of %"PRIu64". Bytes percent %2.2f%%\n",
		   vnode->bytes, node->bytes_size, percent );

	// Not watched enough
	if (percent < visited_percent) return;

	if (vnode->seen) return;

	// Update the DB here.
	debugf("[visited] adding to db '%s'\n", vnode->fname);

	vnode->seen = 1;

	// Write us to disk.
	//visited_db_update(vnode);
}


int visited_file(request_t *node, file_t *file)
{
	visited_t *vnode;

	vnode = visited_find(file->name);
	if (!vnode) return 0;

	debugf("[visited] testing '%s'. %s found.\n", file->name,
		   vnode ? "" : "not");

	return vnode->seen;
}

